/*
 * Delay line
 */


#include <math.h>
#include <string.h>
#include "delay_line.h"

extern double round ( double );


/*
 * Constructor
 * duration in milliseconds, sound rate in Hz
 */
DelayLine *
delay_line_new ( gdouble duration,
                 const guint sound_rate )
{
  DelayLine *ret;
  gint i;

  ret = (DelayLine *) g_malloc ( sizeof(DelayLine) );

  ret->duration = duration;
  ret->dsize = ceil ( (sound_rate*duration)/1000.0 );
  ret->size = (gint) ret->dsize;
  ret->idx = 0;
  ret->line = (gdouble *) g_malloc ( sizeof(gdouble) * ret->size );

  for ( i=0; i<ret->size; ++i )
    ret->line[i] = 0.0;

  return ret;
}


/*
 * Cloning operation
 */
DelayLine *
delay_line_clone ( DelayLine *dl )
{
  DelayLine *ret;

  ret = (DelayLine *) g_malloc ( sizeof(DelayLine) );

  memcpy ( ret, dl, sizeof(DelayLine) );
  ret->line = (gdouble *) g_malloc ( sizeof(gdouble) * ret->size );
  memcpy ( ret->line, dl->line, (sizeof(gdouble)*ret->size) );

  return ret;
}


/*
 * Getting the size of the delay line
 */
gint
delay_line_get_size ( DelayLine *dl )
{
  g_assert ( dl != NULL );

  return dl->size;
}


/*
 * Feeding a delay line
 */
void
delay_line_feed ( DelayLine *dl,
                  gdouble in )
{
  g_assert ( dl != NULL );

  dl->line[dl->idx] = in;
  dl->idx = (dl->idx+1) % dl->size;
}


/*
 * Reading a value in a delay line
 */
static gdouble
linear_interpolation ( gdouble y1,
                       gdouble y2,
                       gdouble mu )
{
  return ( y1*(1.0-mu) + (y2*mu) );
}


gdouble
delay_line_read ( DelayLine *dl,
                  gdouble delay )
{
  gint k1;
  gint k2;
  gdouble v;

  g_assert ( dl != NULL );
  g_assert ( delay <= dl->duration );

  v = (delay/dl->duration) * dl->dsize;
  k1 = dl->idx - (gint) floor(v);
  k2 = k1 + 1;
  v = v - floor(v);

  if ( k1 < 0 )
    k1 += dl->size;

  if ( k2 < 0 )
    k2 += dl->size;

  return linear_interpolation(dl->line[k1], dl->line[k2], v );
}


/*
 * Reading the oldest value in a delay line
 */
gdouble
delay_line_read_oldest ( DelayLine *dl )
{
  g_assert ( dl != NULL );

  return dl->line[dl->idx];
}


/*
 * Reset the delay line
 */
void
delay_line_reset ( DelayLine *dl )
{
  gint i;

  g_assert ( dl != NULL );

  for ( i=0; i<dl->size; ++i )
    dl->line[i] = 0.0;

  dl->idx = 0;
}


/*
 * Destructor
 */
void
delay_line_destroy ( DelayLine *dl )
{
  g_assert ( dl != NULL );

  g_free ( dl->line );
  g_free ( dl );
}
