/*
 * Compressor Effect
 */

#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include "effect.h"
#include "wave.h"
#include "delay_line.h"

extern guint SOUND_RATE;
extern WaveGeneratorsLibrary *gen_lib;
extern EffectsLibrary *effects_lib;

#define ATTACK_MAX  100
#define RELEASE_MAX 2000


/*
 * Internal data structure
 */
typedef struct st_fx_compressor_data
{
  gdouble attack;
  gdouble release;
  gdouble threshold;

  DelayLine *rms_delay_line;
  gint rms_delay_line_size;
  gdouble rms_intermed;

  GtkVBox *vbox;
  GtkHBox *hbox_gen;
  GtkComboBox *combo_gen;

} EffectCompressorData ;


/*
 * Initilization of the compressor effect.
 */
void effect_compressor_init ( gpointer data )
{
  EffectCompressorData *ecd;

  if ( data == NULL )
    return;

  ecd = (EffectCompressorData *) data;

  ecd->attack = 50.0;
  ecd->release = 250.0;
  ecd->threshold = 0.5;

  ecd->rms_delay_line = delay_line_new ( 2.0, SOUND_RATE ); /* 2ms rms buffer (may not be good)*/
  ecd->rms_delay_line_size = delay_line_get_size ( ecd->rms_delay_line );
  ecd->rms_intermed = 0.0;

  ecd->vbox = NULL;
}


/*
 * Cloning operation.
 * Returns a brand new sine wave generator.
 */
gpointer
effect_compressor_clone ( gpointer data )
{
  Effect * ret = g_malloc ( sizeof(Effect) );

  memcpy ( ret, &EffectCompressor, sizeof(Effect) );

  ret->data = g_malloc ( sizeof(EffectCompressorData) );

  if ( data != NULL )
    {
      memcpy ( ret->data, data, sizeof(EffectCompressorData) );
      ((EffectCompressorData*)ret->data)->rms_delay_line = delay_line_clone ( ((EffectCompressorData*)data)->rms_delay_line );
    }
  else
    effect_compressor_init ( ret->data );

  ((EffectCompressorData*)ret->data)->vbox = NULL;

  return ret;
}


/*
 * Makes the configuration user interface
 */
static void
on_attack_changed ( GtkSpinButton *sb,
                   gpointer data )
{
  g_assert ( data != NULL );

  ((EffectCompressorData *) data)->attack = gtk_spin_button_get_value ( sb );
}


static void
on_release_changed ( GtkSpinButton *sb,
                   gpointer data )
{
  g_assert ( data != NULL );

  ((EffectCompressorData *) data)->release = gtk_spin_button_get_value ( sb );
}

static void
on_threshold_changed ( GtkSpinButton *sb,
                       gpointer data )
{
  g_assert ( data != NULL );

  ((EffectCompressorData *) data)->threshold = gtk_spin_button_get_value ( sb );
}


void
effect_compressor_make_ui ( gpointer data,
                          GtkBox *parent )
{
  GtkWidget *widget;
  GtkWidget *box;
  EffectCompressorData *ecd;

  g_assert ( data != NULL );

  ecd = (EffectCompressorData *) data;

  ecd->vbox = GTK_VBOX ( gtk_vbox_new(FALSE,5) );
  gtk_box_pack_start ( parent, GTK_WIDGET(ecd->vbox), TRUE, TRUE, 0 );

  /* attack */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(ecd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "attack (ms) :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  widget = gtk_spin_button_new_with_range ( 1.0, ATTACK_MAX, 1.0 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), ecd->attack );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_attack_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  /* release */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(ecd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "release (ms) :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  widget = gtk_spin_button_new_with_range ( 0.0, RELEASE_MAX, 1.0 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), ecd->release );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_release_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  /* threshold */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(ecd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "threshold :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  widget = gtk_spin_button_new_with_range ( 0.0, 1.0, 0.05 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), ecd->threshold );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_threshold_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );
}


/*
 * Shows/Hides the configuration user interface
 */
void
effect_compressor_show_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_show ( GTK_WIDGET(((EffectCompressorData *)data)->vbox) );
}

void
effect_compressor_hide_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_hide ( GTK_WIDGET(((EffectCompressorData *)data)->vbox) );
}


/*
 * Applies the effect.
 * Returns the new value.
 */
static gdouble
rms ( EffectCompressorData *ecd,
      gdouble in )
{
  gdouble oldest2;
  gdouble in2;

  in2 = in * in;

  oldest2 = delay_line_read_oldest ( ecd->rms_delay_line );
  delay_line_feed ( ecd->rms_delay_line, in2 ); /* values are inserted squared */

  ecd->rms_intermed -= oldest2;
  ecd->rms_intermed += in2;

  return sqrt ( ecd->rms_intermed / ecd->rms_delay_line_size );
}

gdouble
effect_compressor_apply ( gpointer data, gdouble in )
{
  EffectCompressorData *ecd;
  gdouble rms_val;

  ecd = (EffectCompressorData *) data;
  rms_val = rms ( ecd, in );

  if ( rms_val < ecd->threshold )
    rms_val  = 0.0;

  return in - (in * rms_val * rms_val);
}


/*
 * Resets the effect.
 */
void
effect_compressor_reset ( gpointer data )
{
  EffectCompressorData *ecd;

  g_assert ( data != NULL );

  ecd = (EffectCompressorData *) data;

  delay_line_reset ( ecd->rms_delay_line );
  ecd->rms_intermed = 0.0;
}


/*
 * Destroys the effect.
 */
void
effect_compressor_destroy ( gpointer data )
{
  EffectCompressorData *ecd;

  g_assert ( data != NULL );

  ecd = (EffectCompressorData *) data;

  if ( ecd->vbox != NULL )
    gtk_widget_destroy ( GTK_WIDGET(ecd->vbox) );

  delay_line_destroy ( ecd->rms_delay_line );

  g_free ( data );
}


/*
 * Compressor Effect
 */
Effect EffectCompressor = 
  {
    "compressor",
    TRUE,
    NULL,
    effect_compressor_init,
    effect_compressor_clone,
    effect_compressor_make_ui,
    effect_compressor_show_ui,
    effect_compressor_hide_ui,
    effect_compressor_apply,
    effect_compressor_reset,
    effect_compressor_destroy
  } ;
