/*
 * Flanger Effect
 */

#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include "effect.h"
#include "wave.h"
#include "delay_line.h"

extern guint SOUND_RATE;
extern WaveGeneratorsLibrary *gen_lib;
extern EffectsLibrary *effects_lib;

#define DELAY_MAX       10.0
#define SWEEP_DEPTH_MAX 10.0


/*
 * Internal data structure
 */
typedef struct st_fx_flanger_data
{
  gdouble depth;
  gdouble delay;
  gdouble sweep_depth;
  Wave *input;

  DelayLine *delay_line;

  GtkVBox *vbox;
  GtkHBox *hbox_gen;
  GtkComboBox *combo_gen;

} EffectFlangerData ;


/*
 * Initilization of the flanger effect.
 */
void effect_flanger_init ( gpointer data )
{
  EffectFlangerData *efd;

  if ( data == NULL )
    return;

  efd = (EffectFlangerData *) data;

  efd->depth = 1.0;
  efd->delay = 5.0;
  efd->sweep_depth = 2.0;
  efd->delay_line = delay_line_new ( DELAY_MAX+SWEEP_DEPTH_MAX, SOUND_RATE );

  efd->input = wave_new ( gen_lib, effects_lib );

  efd->vbox = NULL;
}


/*
 * Cloning operation.
 * Returns a brand new sine wave generator.
 */
gpointer
effect_flanger_clone ( gpointer data )
{
  Effect * ret = g_malloc ( sizeof(Effect) );

  memcpy ( ret, &EffectFlanger, sizeof(Effect) );

  ret->data = g_malloc ( sizeof(EffectFlangerData) );

  if ( data != NULL )
    {
      memcpy ( ret->data, data, sizeof(EffectFlangerData) );
      ((EffectFlangerData*)ret->data)->input = wave_clone ( ((EffectFlangerData*)data)->input );
      ((EffectFlangerData*)ret->data)->delay_line = delay_line_clone ( ((EffectFlangerData*)data)->delay_line );
    }
  else
    effect_flanger_init ( ret->data );

  ((EffectFlangerData*)ret->data)->vbox = NULL;
  ((EffectFlangerData*)ret->data)->hbox_gen = NULL;
  ((EffectFlangerData*)ret->data)->combo_gen = NULL;

  return ret;
}


/*
 * Makes the configuration user interface
 */
static void
add_wave_gen_to_combo ( gpointer wg_name,
                        gpointer wg,
                        gpointer combo )
{
  gtk_combo_box_append_text ( GTK_COMBO_BOX(combo), wg_name );
}

static void
on_wave_generator_changed ( GtkComboBox *combo,
                            EffectFlangerData *efd )
{
  WaveGenerator *gen;

  gen = wave_generators_library_get_new ( efd->input->gen_lib, gtk_combo_box_get_active_text(combo) );

  if ( efd->input->gen != NULL )
    {
      efd->input->gen->hide_ui ( efd->input->gen->data );
      wave_generator_destroy ( efd->input->gen );
    }

  if ( gen != NULL )
    {
      gen->make_ui ( gen->data, GTK_BOX(efd->hbox_gen) );
      gen->show_ui ( gen->data );
    }

  g_mutex_lock ( efd->input->mutex );
  efd->input->gen = gen;
  g_mutex_unlock ( efd->input->mutex );
}

static void
on_delay_changed ( GtkSpinButton *sb,
                   gpointer data )
{
  EffectFlangerData *efd;

  g_assert ( data != NULL );

  efd = (EffectFlangerData *) data;
  efd->delay = gtk_spin_button_get_value ( sb );
}

static void
on_sweep_depth_changed ( GtkSpinButton *sb,
                   gpointer data )
{
  g_assert ( data != NULL );

  ((EffectFlangerData *) data)->sweep_depth = gtk_spin_button_get_value ( sb );
}

static void
on_depth_changed ( GtkSpinButton *sb,
                   gpointer data )
{
  g_assert ( data != NULL );

  ((EffectFlangerData *) data)->depth = gtk_spin_button_get_value ( sb );
}

void
effect_flanger_make_ui ( gpointer data,
                          GtkBox *parent )
{
  GtkWidget *widget;
  GtkWidget *box;
  EffectFlangerData *efd;

  g_assert ( data != NULL );

  efd = (EffectFlangerData *) data;

  efd->vbox = GTK_VBOX ( gtk_vbox_new(FALSE,5) );
  gtk_box_pack_start ( parent, GTK_WIDGET(efd->vbox), TRUE, TRUE, 0 );

  /* input wave UI */
  efd->hbox_gen = GTK_HBOX ( gtk_hbox_new(FALSE,5) );
  gtk_box_pack_start ( GTK_BOX(efd->vbox), GTK_WIDGET(efd->hbox_gen), FALSE, TRUE, 0 );
  gtk_widget_show ( GTK_WIDGET(efd->hbox_gen) );

  efd->combo_gen = GTK_COMBO_BOX ( gtk_combo_box_new_text() );
  wave_generators_library_foreach ( efd->input->gen_lib, add_wave_gen_to_combo, efd->combo_gen );
  gtk_box_pack_start ( GTK_BOX(efd->hbox_gen), GTK_WIDGET(efd->combo_gen), FALSE, FALSE, 0 );
  gtk_widget_show ( GTK_WIDGET(efd->combo_gen) );

  if ( efd->input->gen != NULL )
    {
      int i;
      gchar *txt;

      gtk_combo_box_set_active ( efd->combo_gen, 0 );
      txt = gtk_combo_box_get_active_text ( efd->combo_gen );
      i = 1;

      while ( txt != NULL )
        {
          if ( strcmp(txt,efd->input->gen->name) == 0 )
            break;

          gtk_combo_box_set_active ( efd->combo_gen, i++ );
          txt = gtk_combo_box_get_active_text ( efd->combo_gen );
        }
    }

  g_signal_connect ( efd->combo_gen, "changed", G_CALLBACK(on_wave_generator_changed), efd );

  widget = gtk_vseparator_new ( );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(efd->hbox_gen), widget, FALSE, TRUE, 0 );

  /* delay */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(efd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "delay (ms) :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  widget = gtk_spin_button_new_with_range ( 0.0, DELAY_MAX, 0.5 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), efd->delay );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_delay_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  /* sweep depth */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(efd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "sweep depth (ms) :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  widget = gtk_spin_button_new_with_range ( 0.0, SWEEP_DEPTH_MAX, 0.5 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), efd->sweep_depth );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_sweep_depth_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  /* depth */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(efd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "depth :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );

  widget = gtk_spin_button_new_with_range ( 0.0, 1.0, 0.05 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), efd->depth );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_depth_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );
}


/*
 * Shows/Hides the configuration user interface
 */
void
effect_flanger_show_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_show ( GTK_WIDGET(((EffectFlangerData *)data)->vbox) );
}

void
effect_flanger_hide_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_hide ( GTK_WIDGET(((EffectFlangerData *)data)->vbox) );
}


/*
 * Applies the effect.
 * Returns the new value.
 */
gdouble
effect_flanger_apply ( gpointer data, gdouble in )
{
  EffectFlangerData *efd;
  gdouble last;
  gdouble sweep;

  g_assert ( data != NULL );

  efd = (EffectFlangerData *) data;

  delay_line_feed ( efd->delay_line, in );

  sweep = (efd->sweep_depth * (1.0 + wave_iter(efd->input))) / 2.0;
  last = delay_line_read ( efd->delay_line, efd->delay+sweep );
    
  return (in+(last*efd->depth)) / (1.0+efd->depth);
}


/*
 * Resets the effect.
 */
void
effect_flanger_reset ( gpointer data )
{
  EffectFlangerData *efd;

  g_assert ( data != NULL );

  efd = (EffectFlangerData *) data;

  if ( efd->input != NULL )
    wave_reset ( efd->input );

  delay_line_reset ( efd->delay_line );
}


/*
 * Destroys the effect.
 */
void
effect_flanger_destroy ( gpointer data )
{
  EffectFlangerData *efd;

  g_assert ( data != NULL );

  efd = (EffectFlangerData *) data;

  if ( efd->input != NULL )
    wave_destroy ( efd->input );

  if ( efd->vbox != NULL )
    gtk_widget_destroy ( GTK_WIDGET(efd->vbox) );

  delay_line_destroy ( efd->delay_line );

  g_free ( data );
}


/*
 * Flanger Effect
 */
Effect EffectFlanger = 
  {
    "flanger",
    TRUE,
    NULL,
    effect_flanger_init,
    effect_flanger_clone,
    effect_flanger_make_ui,
    effect_flanger_show_ui,
    effect_flanger_hide_ui,
    effect_flanger_apply,
    effect_flanger_reset,
    effect_flanger_destroy
  } ;
