/*
 * Ring Mod Effect
 */

#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include "effect.h"
#include "wave.h"

extern guint SOUND_RATE;
extern WaveGeneratorsLibrary *gen_lib;
extern EffectsLibrary *effects_lib;


/*
 * Internal data structure
 */
typedef struct st_fx_ring_mod_data
{
  Wave *input;

  GtkVBox *vbox;
  GtkHBox *hbox_gen;
  GtkComboBox *combo_gen;

} EffectRingModData ;


/*
 * Initilization of the ring mod effect.
 */
void effect_ring_mod_init ( gpointer data )
{
  EffectRingModData *ermd;

  if ( data == NULL )
    return;

  ermd = (EffectRingModData *) data;

  ermd->input = wave_new ( gen_lib, effects_lib );
  ermd->vbox = NULL;
}


/*
 * Cloning operation.
 * Returns a brand new sine wave generator.
 */
gpointer
effect_ring_mod_clone ( gpointer data )
{
  Effect * ret = g_malloc ( sizeof(Effect) );

  memcpy ( ret, &EffectRingMod, sizeof(Effect) );

  ret->data = g_malloc ( sizeof(EffectRingModData) );

  if ( data != NULL )
    {
      memcpy ( ret->data, data, sizeof(EffectRingModData) );
      ((EffectRingModData*)ret->data)->input = wave_clone ( ((EffectRingModData*)data)->input );
    }
  else
    effect_ring_mod_init ( ret->data );

  ((EffectRingModData*)ret->data)->vbox = NULL;
  ((EffectRingModData*)ret->data)->hbox_gen = NULL;
  ((EffectRingModData*)ret->data)->combo_gen = NULL;

  return ret;
}


/*
 * Makes the configuration user interface
 */
static void
add_wave_gen_to_combo ( gpointer wg_name,
                        gpointer wg,
                        gpointer combo )
{
  gtk_combo_box_append_text ( GTK_COMBO_BOX(combo), wg_name );
}

static void
on_wave_generator_changed ( GtkComboBox *combo,
                            EffectRingModData *ermd )
{
  WaveGenerator *gen;

  gen = wave_generators_library_get_new ( ermd->input->gen_lib, gtk_combo_box_get_active_text(combo) );

  if ( ermd->input->gen != NULL )
    {
      ermd->input->gen->hide_ui ( ermd->input->gen->data );
      wave_generator_destroy ( ermd->input->gen );
    }

  if ( gen != NULL )
    {
      gen->make_ui ( gen->data, GTK_BOX(ermd->hbox_gen) );
      gen->show_ui ( gen->data );
    }

  g_mutex_lock ( ermd->input->mutex );
  ermd->input->gen = gen;
  g_mutex_unlock ( ermd->input->mutex );
}

void
effect_ring_mod_make_ui ( gpointer data,
                          GtkBox *parent )
{
  GtkWidget *widget;
  EffectRingModData *ermd;

  g_assert ( data != NULL );

  ermd = (EffectRingModData *) data;

  ermd->vbox = GTK_VBOX ( gtk_vbox_new(FALSE,5) );
  gtk_box_pack_start ( parent, GTK_WIDGET(ermd->vbox), TRUE, TRUE, 0 );

  /* input wave UI */
  ermd->hbox_gen = GTK_HBOX ( gtk_hbox_new(FALSE,5) );
  gtk_box_pack_start ( GTK_BOX(ermd->vbox), GTK_WIDGET(ermd->hbox_gen), FALSE, TRUE, 0 );
  gtk_widget_show ( GTK_WIDGET(ermd->hbox_gen) );

  ermd->combo_gen = GTK_COMBO_BOX ( gtk_combo_box_new_text() );
  wave_generators_library_foreach ( ermd->input->gen_lib, add_wave_gen_to_combo, ermd->combo_gen );
  gtk_box_pack_start ( GTK_BOX(ermd->hbox_gen), GTK_WIDGET(ermd->combo_gen), FALSE, FALSE, 0 );
  gtk_widget_show ( GTK_WIDGET(ermd->combo_gen) );

  if ( ermd->input->gen != NULL )
    {
      int i;
      gchar *txt;

      gtk_combo_box_set_active ( ermd->combo_gen, 0 );
      txt = gtk_combo_box_get_active_text ( ermd->combo_gen );
      i = 1;

      while ( txt != NULL )
        {
          if ( strcmp(txt,ermd->input->gen->name) == 0 )
            break;

          gtk_combo_box_set_active ( ermd->combo_gen, i++ );
          txt = gtk_combo_box_get_active_text ( ermd->combo_gen );
        }
    }

  g_signal_connect ( ermd->combo_gen, "changed", G_CALLBACK(on_wave_generator_changed), ermd );

  widget = gtk_vseparator_new ( );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(ermd->hbox_gen), widget, FALSE, TRUE, 0 );
}


/*
 * Shows/Hides the configuration user interface
 */
void
effect_ring_mod_show_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_show ( GTK_WIDGET(((EffectRingModData *)data)->vbox) );
}

void
effect_ring_mod_hide_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_hide ( GTK_WIDGET(((EffectRingModData *)data)->vbox) );
}


/*
 * Applies the effect.
 * Returns the new value.
 */
gdouble
effect_ring_mod_apply ( gpointer data, gdouble in )
{
  g_assert ( data != NULL );

  return in * wave_iter(((EffectRingModData *)data)->input);
}


/*
 * Resets the effect.
 */
void
effect_ring_mod_reset ( gpointer data )
{
  EffectRingModData *ermd;

  g_assert ( data != NULL );

  ermd = (EffectRingModData *) data;

  if ( ermd->input != NULL )
    wave_reset ( ermd->input );
}


/*
 * Destroys the effect.
 */
void
effect_ring_mod_destroy ( gpointer data )
{
  EffectRingModData *ermd;

  g_assert ( data != NULL );

  ermd = (EffectRingModData *) data;

  if ( ermd->input != NULL )
    wave_destroy ( ermd->input );

  if ( ermd->vbox != NULL )
    gtk_widget_destroy ( GTK_WIDGET(ermd->vbox) );

  g_free ( data );
}


/*
 * Ring Mod Effect
 */
Effect EffectRingMod = 
  {
    "ring mod",
    TRUE,
    NULL,
    effect_ring_mod_init,
    effect_ring_mod_clone,
    effect_ring_mod_make_ui,
    effect_ring_mod_show_ui,
    effect_ring_mod_hide_ui,
    effect_ring_mod_apply,
    effect_ring_mod_reset,
    effect_ring_mod_destroy
  } ;
