#include <gtk/gtk.h>
#include <glib/gprintf.h>
#include <glade/glade.h>
#include <math.h>

#include "wave.h"
#include "snd_out/snd_out_alsa.h"

GladeXML *gui_xml;

WaveGeneratorsLibrary *gen_lib;
EffectsLibrary *effects_lib;
Wave *selected_wave = NULL;
SndOutAlsa *sound;

/*
 * Callback for the application desinit.
 */
void
on_quit_activate ( GtkWidget *widget,
                   gpointer data )
{
  snd_out_alsa_destroy ( sound );
  effects_library_destroy ( effects_lib );
  wave_generators_library_destroy ( gen_lib );
  gtk_main_quit ( );
}


/*
 * Callback for playing/stopping of the playback.
 */
void
on_toolbutton_play_stop_clicked ( GtkToolButton *tb,
                                  gpointer data )
{
  if ( selected_wave == NULL )
    return;

  if ( !snd_out_alsa_is_playing(sound) )
    {
      gtk_tool_button_set_stock_id ( tb, GTK_STOCK_MEDIA_STOP );
      wave_reset ( selected_wave );
      snd_out_alsa_play ( sound, selected_wave );
    }
  else
    {
      gtk_tool_button_set_stock_id ( tb, GTK_STOCK_MEDIA_PLAY );
      snd_out_alsa_stop ( sound );
    }
}


/*
 * Callback for plotting of the selected wave.
 */
void
on_toolbutton_plot_clicked ( GtkToolButton *tb,
                             gpointer data )
{
  if ( selected_wave == NULL )
    return;

  gint i;
  FILE *plot_file;
  Wave *w;

  const gint limit = (gint) ((gdouble) SOUND_RATE/10);

  plot_file = fopen ( "wave.plot", "w" );

  w = wave_clone ( selected_wave );
  wave_reset ( w );

  for ( i=0; i<limit; ++i )
    g_fprintf ( plot_file, "%f\n", wave_iter(w) );

  fclose ( plot_file );

  wave_destroy ( w );

  system ( "echo \"plot \'wave.plot\' with lines\" | gnuplot -persist" );
}


/*
 * Callback for the addition of a wave.
 */
#define WAVES_COL_NAME   0
#define WAVES_COL_OBJECT 1
#define NB_COLS_WAVES    2

void
on_add_wave_clicked ( GtkTreeView *tv,
                      gpointer data )
{
  GtkListStore *store;
  GtkTreeIter iter;
  Wave *w;

  store = GTK_LIST_STORE ( gtk_tree_view_get_model(tv) );

  gtk_list_store_append ( store, &iter );

  w = wave_new ( gen_lib, effects_lib );
  wave_make_ui ( w, GTK_BOX(glade_xml_get_widget(gui_xml,"vbox_wave")) );

  gtk_list_store_set ( store, &iter,
                       WAVES_COL_NAME, w->name,
                       WAVES_COL_OBJECT, w, -1 );
}


/*
 * Callback for selection change in the waves list
 */
void
on_selected_wave_changed ( GtkTreeSelection *selection,
                           gpointer data )
{
  GtkTreeView *tree;
  GtkTreeModel *model;
  GtkTreeIter iter;
  Wave *w=NULL;

  tree = gtk_tree_selection_get_tree_view ( selection );

  if ( gtk_tree_selection_get_selected(selection, &model, &iter) )
    gtk_tree_model_get ( model, &iter, WAVES_COL_OBJECT, &w, -1 );

  if ( w == NULL )
    {
      if ( selected_wave != NULL )
        wave_hide_ui ( selected_wave );
    }
  else
    {
      if ( selected_wave != NULL )
        wave_hide_ui ( selected_wave );

      wave_show_ui ( w );
    }

  selected_wave = w;
}


/*
 * Callback for the removal of a wave.
 */
void
on_remove_wave_clicked ( GtkTreeView *tv,
                         gpointer data )
{
  if ( selected_wave == NULL )
    return;

  if ( snd_out_alsa_is_playing(sound) )
    {
      gtk_tool_button_set_stock_id ( GTK_TOOL_BUTTON(glade_xml_get_widget(gui_xml,"toolbutton_play_stop")), GTK_STOCK_MEDIA_PLAY );
      snd_out_alsa_stop ( sound );
    }

  /* todo */
}


/*
 * Callback for the clearing of the waves list.
 */
void
on_clear_waves_list_clicked ( GtkTreeView *tv,
                              gpointer data )
{
  if ( snd_out_alsa_is_playing(sound) )
    {
      gtk_tool_button_set_stock_id ( GTK_TOOL_BUTTON(glade_xml_get_widget(gui_xml,"toolbutton_play_stop")), GTK_STOCK_MEDIA_PLAY );
      snd_out_alsa_stop ( sound );
    }

  /* todo */
}


/*
 * Waves list initialization
 */
void
wave_name_edited ( GtkCellRendererText *renderer,
                   gchar *path,
                   gchar *new_text,
                   GtkTreeSelection *selection )
{
  GtkTreeModel *model;
  GtkTreeIter iter;

  if ( gtk_tree_selection_get_selected(selection, &model, &iter) )
    {
      gtk_list_store_set ( GTK_LIST_STORE(model), &iter, WAVES_COL_NAME, new_text, -1 );
      wave_set_name ( selected_wave, new_text );
    }
}

void
init_waves_list ( void )
{
  GtkTreeView *tree;
  GtkListStore *list_store;
  GtkCellRenderer *renderer;

  tree = GTK_TREE_VIEW ( glade_xml_get_widget(gui_xml,"treeview_waves_list") );
  list_store = gtk_list_store_new ( NB_COLS_WAVES, G_TYPE_STRING, G_TYPE_POINTER );
  gtk_tree_view_set_model ( GTK_TREE_VIEW(tree), GTK_TREE_MODEL(list_store) );

  renderer = gtk_cell_renderer_text_new ( );
  g_object_set ( G_OBJECT(renderer), "editable", TRUE, NULL );
  g_signal_connect ( renderer, "edited", G_CALLBACK(wave_name_edited), gtk_tree_view_get_selection(tree) );
  gtk_tree_view_insert_column_with_attributes ( tree, WAVES_COL_NAME, "name", renderer, "text", WAVES_COL_NAME, NULL );

  gtk_tree_selection_set_mode ( gtk_tree_view_get_selection(tree), GTK_SELECTION_SINGLE );
  g_signal_connect ( gtk_tree_view_get_selection(tree), "changed", G_CALLBACK(on_selected_wave_changed), NULL );

  g_object_unref ( list_store );
}


/*
 * Main
 */
int
main ( int argc,
       char **argv )
{
  /* GLib/GDK thread support */
  if ( !g_thread_supported() )
    g_thread_init ( NULL );

  gdk_threads_init ( );

  gen_lib = wave_generators_library_new ( );
  effects_lib = effects_library_new ( );

  /* GUI init */
  gtk_init ( &argc, &argv );
  gui_xml = glade_xml_new ( "glade/jof.glade", NULL, NULL );

  if ( gui_xml == NULL )
    return -1;

  gtk_about_dialog_set_version ( GTK_ABOUT_DIALOG(glade_xml_get_widget(gui_xml,"win_about")), SVN_REVISION );

  /* connect signal handlers */
  glade_xml_signal_autoconnect ( gui_xml );

  /* build wave list */
  init_waves_list ( );

  /* sound */
  sound = snd_out_alsa_new ( );

  /* main loop */
  gtk_widget_show ( glade_xml_get_widget(gui_xml,"win_main") );
  gtk_main ( );

  return 0;
}
