/*
 * Sound Output with ALSA
 */

#include "snd_out_alsa.h"

guint SOUND_RATE = 44100;

/*
 * Constructors
 */
SndOutAlsa *
snd_out_alsa_new ( void )
{
  SndOutAlsa *ret;
  gint rc, dir;

  ret = (SndOutAlsa *) g_malloc ( sizeof(SndOutAlsa) );

  ret->play = FALSE;

  /* Open PCM device for playback */
  rc = snd_pcm_open ( &ret->handle, "default", SND_PCM_STREAM_PLAYBACK, 0 );
  if ( rc < 0 )
    {
      g_error ( "unable to open pcm device: %s", snd_strerror(rc) );
      return NULL;
    }

  /* Allocate a hardware parameters object */
  snd_pcm_hw_params_alloca ( &ret->params );

  /* Fill it in with default values */
  rc = snd_pcm_hw_params_any ( ret->handle, ret->params );
  if ( rc < 0)
    {
      g_error ( "unable to initialize hw parameters: %s", snd_strerror(rc) );
      return NULL;
    }

  /* Set the desired hardware parameters */
  rc = snd_pcm_hw_params_set_rate_resample ( ret->handle, ret->params, 0 );
	if ( rc < 0 )
    {
      g_error ( "unable to setup resampling: %s\n", snd_strerror(rc) );
      return NULL;
    }

  rc = snd_pcm_hw_params_set_access ( ret->handle, ret->params, SND_PCM_ACCESS_RW_INTERLEAVED ); /* Interleaved mode */
  if ( rc < 0)
    {
      g_error ( "unable to set hw access: %s", snd_strerror(rc) );
      return NULL;
    }

  rc = snd_pcm_hw_params_set_format ( ret->handle, ret->params, SND_PCM_FORMAT_S16_LE );         /* Signed 16-bit little-endian format */
  if ( rc < 0)
    {
      g_error ( "unable to set hw format: %s", snd_strerror(rc) );
      return NULL;
    }

  rc = snd_pcm_hw_params_set_channels ( ret->handle, ret->params, 2 );                           /* Two channels (stereo) */
  if ( rc < 0)
    {
      g_error ( "unable to set hw channels: %s", snd_strerror(rc) );
      return NULL;
    }

  rc = snd_pcm_hw_params_set_rate_near ( ret->handle, ret->params, &SOUND_RATE, &dir );          /* 44100 Hz, sampling rate */
  if ( rc < 0)
    {
      g_error ( "unable to set hw rate: %s", snd_strerror(rc) );
      return NULL;
    }
  /* check SOUND_RATE == 44100 ? */

  /* Set period size to 32 frames */
  ret->snd_frames = 32;
  rc = snd_pcm_hw_params_set_period_size_near ( ret->handle, ret->params, &ret->snd_frames, &dir );
  if ( rc < 0)
    {
      g_error ( "unable to set hw period time: %s", snd_strerror(rc) );
      return NULL;
    }

  /* Write the parameters to the driver */
  rc = snd_pcm_hw_params ( ret->handle, ret->params );
  if ( rc < 0 )
    {
      g_error ( "unable to set hw parameters: %s", snd_strerror(rc) );
      return NULL;
    }

  /* Use a buffer large enough to hold one period */
  snd_pcm_hw_params_get_period_size ( ret->params, &ret->snd_frames, &dir );
  ret->buf_size = ret->snd_frames * 2; /* 2 channels */
  ret->buffer = (gint16 *) g_malloc ( sizeof(gint16) * ret->buf_size );

  snd_pcm_hw_params_get_period_time ( ret->params, &ret->snd_period_time, &dir );

  /* threading stuffs */
  ret->thread = NULL;

  return ret;
}


/*
 * Playback a wave
 */
gpointer
play ( gpointer data )
{
  SndOutAlsa *soa;
  gint16 snd_val;
  gint i, rc;

  soa = (SndOutAlsa *) data;

  snd_pcm_prepare ( soa->handle );

  while ( soa->play )
    {
      for ( i=0; i<soa->buf_size; i+=2 )
        {
          snd_val = ( gint16 )  ( wave_iter(soa->wave) * 32766.0 );
          soa->buffer[i] = snd_val;
          soa->buffer[i+1] = snd_val;
        }

      rc = snd_pcm_writei ( soa->handle, soa->buffer, soa->snd_frames );
      
      if ( rc == -EPIPE ) /* EPIPE means underrun */
        snd_pcm_prepare ( soa->handle );
      else if ( rc < 0 )
        g_warning ( "snd_out_alsa: writei() said: %s", snd_strerror(rc) );
      else if ( rc != (int)soa->snd_frames )
        g_warning ( "snd_out_alsa: short write, wrote %d frames", rc );
    }

  return NULL;
}

void
snd_out_alsa_play ( SndOutAlsa *soa,
                    Wave *w )
{
  g_assert ( soa != NULL );

  if ( soa->thread != NULL )
    return;

  soa->play = TRUE;
  soa->wave = w;
  soa->thread = g_thread_create ( play, soa, TRUE, NULL );
}


/*
 * Is playing ?
 */
gboolean
snd_out_alsa_is_playing ( SndOutAlsa *soa )
{
  g_assert ( soa != NULL );

  return soa->play;
}


/*
 * Stop playback
 */
void
snd_out_alsa_stop ( SndOutAlsa *soa )
{
  g_assert ( soa != NULL );

  if ( soa->thread == NULL )
    return;

  soa->play = FALSE;

  g_thread_join ( soa->thread );
  soa->thread = NULL;
}


/*
 * Destructor
 */
void
snd_out_alsa_destroy ( SndOutAlsa *soa )
{
  g_assert ( soa != NULL );

  if ( soa->thread != NULL )
    snd_out_alsa_stop ( soa );

  snd_pcm_drain ( soa->handle );
  snd_pcm_close ( soa->handle );

  g_free ( soa->buffer );
  g_free ( soa );
}
