/*
 * Wave
 */

#include <gtk/gtk.h>
#include <string.h>

#include "wave.h"
#include "wave_generators_library.h"


static gint w_index=0;

/*
 * Constructors
 * Return a new wave.
 */
Wave *
wave_new ( WaveGeneratorsLibrary *gen_lib,
           EffectsLibrary *effects_lib )
{
  gchar *name;
  Wave *ret;

  name = g_strdup_printf ( "unnamed-%d", w_index++ );
  ret = wave_new_with_name ( gen_lib, effects_lib, name );
  g_free ( name );

  return ret;
}


Wave *
wave_new_with_name ( WaveGeneratorsLibrary *gen_lib,
                     EffectsLibrary *effects_lib,
                     const gchar *name )
{
  Wave *ret;

  ret = (Wave *) g_malloc ( sizeof(Wave) );
  ret->name = g_strdup ( name );
  ret->gen_lib = gen_lib;
  ret->gen = NULL;
  ret->effects = NULL;
  ret->effects_lib = effects_lib;
  ret->selected_effect = NULL;

  ret->generators = NULL;
  ret->vbox = NULL;

  ret->mutex = g_mutex_new ( );

  return ret;
}


/*
 * Sets the name of a wave
 */
void
wave_set_name ( Wave *w,
                const gchar *name )
{
  g_assert ( w != NULL );

  g_free ( w->name );
  w->name = g_strdup ( name );
}


/*
 * Clones a wave
 */
Wave *
wave_clone ( Wave *w )
{
  Wave *ret;
  GSList *l;
  Effect *e;

  ret = (Wave *) g_malloc ( sizeof(Wave) );
  ret->name = g_strdup ( w->name );
  ret->gen_lib = w->gen_lib;

  if ( w->gen != NULL )
    ret->gen = w->gen->clone ( w->gen->data );
  else
    ret->gen = NULL;

  ret->selected_effect = NULL;
  ret->effects = NULL;
  ret->effects_lib = w->effects_lib;

  for ( l=w->effects; l!=NULL; l=g_slist_next(l) )
    {
      e = (Effect *) l->data;
      ret->effects = g_slist_append ( ret->effects, e->clone(e->data) );
    }

  ret->generators = NULL;
  ret->vbox = NULL;

  ret->mutex = g_mutex_new ( );

  return ret;
}


/*
 * Builds the user interface for wave configuration
 */
static void
add_wave_gen_to_combo ( gpointer wg_name,
                        gpointer wg,
                        gpointer combo )
{
  gtk_combo_box_append_text ( GTK_COMBO_BOX(combo), wg_name );
}

static void
on_wave_generator_changed ( GtkComboBox *combo,
                            Wave *w )
{
  WaveGenerator *gen;

  gen = g_hash_table_lookup ( w->generators, gtk_combo_box_get_active_text(combo) );

  if ( gen == NULL )
    {
      gen = wave_generators_library_get_new ( w->gen_lib, gtk_combo_box_get_active_text(combo) );

      if ( gen != NULL )
        {
          gen->make_ui ( gen->data, GTK_BOX(w->hbox_gen) );
          g_hash_table_insert ( w->generators, gen->name, gen );
        }
    }

  if ( w->gen != NULL )
    w->gen->hide_ui ( w->gen->data );

  if ( gen != NULL )
    gen->show_ui ( gen->data );

  g_mutex_lock ( w->mutex );
  w->gen = gen;
  g_mutex_unlock ( w->mutex );
}

#define EFFECTS_COL_NAME   0
#define EFFECTS_COL_OBJECT 1
#define NB_COLS_EFFECTS    2

static void
on_selected_effect_changed ( GtkTreeSelection *selection,
                             gpointer data )
{
  GtkTreeView *tree;
  GtkTreeModel *model;
  GtkTreeIter iter;
  Effect *e=NULL;
  Wave *w;

  w = (Wave *) data;
  tree = gtk_tree_selection_get_tree_view ( selection );

  if ( gtk_tree_selection_get_selected(selection, &model, &iter) )
    gtk_tree_model_get ( model, &iter, EFFECTS_COL_OBJECT, &e, -1 );

  if ( e == NULL )
    {
      if ( w->selected_effect != NULL )
        w->selected_effect->hide_ui ( w->selected_effect->data );
    }
  else
    {
      if ( w->selected_effect != NULL )
        w->selected_effect->hide_ui ( w->selected_effect->data );

      e->show_ui ( e->data );
    }

  w->selected_effect = e;
}

static void
on_effect_chooser_selection_changed ( GtkTreeSelection *selection,
                                      gpointer *data )
{
  GtkTreeView *tree;
  GtkTreeModel *model;
  GtkTreeIter iter;
  
  tree = gtk_tree_selection_get_tree_view ( selection );

  *data = NULL;

  if ( gtk_tree_selection_get_selected(selection, &model, &iter) )
    gtk_tree_model_get ( model, &iter, EFFECTS_COL_OBJECT, data, -1 );
}

static void
add_effect_to_tree ( gpointer effect_name,
                     gpointer effect,
                     gpointer list_store )
{
  GtkTreeIter iter;

  gtk_list_store_append ( list_store, &iter );
  gtk_list_store_set ( list_store, &iter, EFFECTS_COL_NAME, effect_name, EFFECTS_COL_OBJECT, effect, -1 );
}

static Effect *
open_effect_chooser ( Wave *w )
{
  GtkWidget *dialog;
  GtkWidget *scrolledwindow;
  GtkListStore *list_store;
  GtkTreeView *tree;
  GtkTreeSelection *selection;
  GtkCellRenderer *renderer;
  gint response;
  Effect *ret;

  ret = NULL;

  dialog = gtk_dialog_new_with_buttons ( "Choose an effect, please",
                                         NULL, /* should use win main */
                                         GTK_DIALOG_MODAL|GTK_DIALOG_NO_SEPARATOR,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                         GTK_STOCK_OK, GTK_RESPONSE_OK, NULL );

  gtk_window_set_resizable ( GTK_WINDOW(dialog), FALSE );

  /* add the tree */
  scrolledwindow = gtk_scrolled_window_new ( NULL, NULL );
  gtk_widget_show ( scrolledwindow );
  gtk_container_set_border_width ( GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5 );
  gtk_box_pack_start ( GTK_BOX(GTK_DIALOG(dialog)->vbox), scrolledwindow, TRUE, TRUE, 5 );
  gtk_scrolled_window_set_policy ( GTK_SCROLLED_WINDOW (scrolledwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
  gtk_scrolled_window_set_shadow_type ( GTK_SCROLLED_WINDOW (scrolledwindow), GTK_SHADOW_IN );

  tree = GTK_TREE_VIEW ( gtk_tree_view_new() );
  gtk_widget_show ( GTK_WIDGET(tree) );
  gtk_container_add ( GTK_CONTAINER(scrolledwindow), GTK_WIDGET(tree) );
  gtk_tree_view_set_headers_visible ( tree, FALSE );
  gtk_tree_view_set_fixed_height_mode ( tree, TRUE );

  list_store = gtk_list_store_new ( NB_COLS_EFFECTS, G_TYPE_STRING, G_TYPE_POINTER );
  gtk_tree_view_set_model ( tree, GTK_TREE_MODEL(list_store) );

  effects_library_foreach ( w->effects_lib, add_effect_to_tree, list_store );

  renderer = gtk_cell_renderer_text_new ( );
  gtk_tree_view_insert_column_with_attributes ( tree, EFFECTS_COL_NAME, "name", renderer, "text", EFFECTS_COL_NAME, NULL );

  selection = gtk_tree_view_get_selection ( tree );
  gtk_tree_selection_set_mode ( selection, GTK_SELECTION_SINGLE );
  gtk_tree_selection_unselect_all ( selection );
  g_signal_connect ( selection, "changed", G_CALLBACK(on_effect_chooser_selection_changed), &ret );

  g_object_unref ( list_store );

  /* run */
  do
    {
      response = gtk_dialog_run ( GTK_DIALOG(dialog) );
    }
  while ( ( response == GTK_RESPONSE_OK ) && ( ret == NULL ) );

  gtk_widget_destroy ( dialog );

  if ( response != GTK_RESPONSE_OK )
    return NULL;

  return ret;
}

static void
on_add_effect_clicked ( GtkButton *btn,
                        gpointer data )
{
  GtkListStore *store;
  GtkTreeIter iter;
  Wave *w;
  Effect *e;

  w = (Wave *) data;

  store = GTK_LIST_STORE ( gtk_tree_view_get_model(w->tree_effects) );

  e = open_effect_chooser ( w );

  if ( e != NULL )
    {
      Effect *e_instance = e->clone ( e->data );
      e_instance->make_ui ( e_instance->data, GTK_BOX(w->hbox_effects) );

      gtk_list_store_append ( store, &iter );

      gtk_list_store_set ( store, &iter,
                           EFFECTS_COL_NAME, e_instance->name,
                           EFFECTS_COL_OBJECT, e_instance, -1 );

      w->effects = g_slist_append ( w->effects, e_instance );
    }
}

static void
on_remove_effect_clicked ( GtkButton *btn,
                           gpointer data )
{
  GtkTreeSelection *selection;
  GtkListStore *store;
  GtkTreeIter iter;
  Wave *w;
  Effect *e;

  w = (Wave *) data;

  store = GTK_LIST_STORE ( gtk_tree_view_get_model(w->tree_effects) );

  selection = gtk_tree_view_get_selection ( w->tree_effects );

  if ( gtk_tree_selection_get_selected(selection, NULL, &iter) )
    {
      gtk_tree_model_get ( GTK_TREE_MODEL(store), &iter, EFFECTS_COL_OBJECT, &e, -1 );

      gtk_list_store_remove ( store, &iter );
      w->effects = g_slist_remove ( w->effects, e );
      effect_destroy ( e );
      w->selected_effect = NULL;
    }
}

static void
on_clear_effects_list_clicked ( GtkButton *btn,
                                gpointer data )
{
  GtkListStore *store;
  Wave *w;
  GSList *l;

  w = (Wave *) data;

  store = GTK_LIST_STORE ( gtk_tree_view_get_model(w->tree_effects) );
  gtk_list_store_clear ( store );

  for ( l=w->effects; l!=NULL; l=g_slist_next(l) )
    effect_destroy ( (Effect *) l->data );

  g_slist_free ( w->effects );
  w->effects = NULL;
}

void
wave_make_ui ( Wave *w,
               GtkBox *parent )
{
  GtkWidget *widget;
  GtkListStore *list_store;
  GtkCellRenderer *renderer;
  GtkBox *box;
  GtkHBox *button_box;
  GtkButton *btn;

  g_assert ( w != NULL );

  w->vbox = GTK_VBOX ( gtk_vbox_new(FALSE,5) );
  gtk_box_pack_start ( parent, GTK_WIDGET(w->vbox), TRUE, TRUE, 0 );

  widget = gtk_hseparator_new ( );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(w->vbox), widget, FALSE, TRUE, 0 );

   /* wave generator UI */
  w->generators = g_hash_table_new_full ( g_str_hash, g_str_equal, NULL, (GDestroyNotify)wave_generator_destroy );
  widget = gtk_label_new ( "input" );
  gtk_label_set_markup ( GTK_LABEL(widget), "<b>input</b>" );
  gtk_label_set_use_markup ( GTK_LABEL(widget), TRUE );
  gtk_misc_set_alignment ( GTK_MISC(widget), 0.0, 0.0 );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(w->vbox), widget, FALSE, TRUE, 0 );

  w->hbox_gen = GTK_HBOX ( gtk_hbox_new(FALSE,5) );
  gtk_box_pack_start ( GTK_BOX(w->vbox), GTK_WIDGET(w->hbox_gen), FALSE, TRUE, 0 );
  gtk_widget_show ( GTK_WIDGET(w->hbox_gen) );

  w->combo_gen = GTK_COMBO_BOX ( gtk_combo_box_new_text() );
  wave_generators_library_foreach ( w->gen_lib, add_wave_gen_to_combo, w->combo_gen );
  gtk_box_pack_start ( GTK_BOX(w->hbox_gen), GTK_WIDGET(w->combo_gen), FALSE, FALSE, 0 );
  gtk_widget_show ( GTK_WIDGET(w->combo_gen) );

  if ( w->gen != NULL )
    {
      int i;
      gchar *txt;

      gtk_combo_box_set_active ( w->combo_gen, 0 );
      txt = gtk_combo_box_get_active_text ( w->combo_gen );
      i = 1;

      while ( txt != NULL )
        {
          if ( strcmp(txt,w->gen->name) == 0 )
            break;

          gtk_combo_box_set_active ( w->combo_gen, i++ );
          txt = gtk_combo_box_get_active_text ( w->combo_gen );
        }
    }

  g_signal_connect ( w->combo_gen, "changed", G_CALLBACK(on_wave_generator_changed), w );

  widget = gtk_vseparator_new ( );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(w->hbox_gen), widget, FALSE, TRUE, 0 );

  widget = gtk_hseparator_new ( );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(w->vbox), widget, FALSE, TRUE, 0 );

  /* effects UI */
  widget = gtk_label_new ( "effects" );
  gtk_label_set_markup ( GTK_LABEL(widget), "<b>effects</b>" );
  gtk_label_set_use_markup ( GTK_LABEL(widget), TRUE );
  gtk_misc_set_alignment ( GTK_MISC(widget), 0.0, 0.0 );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(w->vbox), widget, FALSE, TRUE, 0 );

  w->hbox_effects = GTK_HBOX ( gtk_hbox_new(FALSE,5) );
  gtk_box_pack_start ( GTK_BOX(w->vbox), GTK_WIDGET(w->hbox_effects), TRUE, TRUE, 0 );
  gtk_widget_show ( GTK_WIDGET(w->hbox_effects) );

  box = GTK_BOX ( gtk_vbox_new(FALSE,5) );
  gtk_widget_show ( GTK_WIDGET(box) );
  gtk_box_pack_start ( GTK_BOX(w->hbox_effects), GTK_WIDGET(box), TRUE, TRUE, 0 );

  w->tree_effects = GTK_TREE_VIEW( gtk_tree_view_new() );
  gtk_tree_view_set_headers_visible ( w->tree_effects, FALSE );
  gtk_tree_view_set_fixed_height_mode ( w->tree_effects, TRUE );
  gtk_tree_view_set_reorderable ( w->tree_effects, FALSE ); /* should be handled */
  gtk_box_pack_start ( box, GTK_WIDGET(w->tree_effects), TRUE, TRUE, 0 );
  gtk_widget_show ( GTK_WIDGET(w->tree_effects) );

  list_store = gtk_list_store_new ( NB_COLS_EFFECTS, G_TYPE_STRING, G_TYPE_POINTER );
  gtk_tree_view_set_model ( w->tree_effects, GTK_TREE_MODEL(list_store) );

  renderer = gtk_cell_renderer_text_new ( );
  gtk_tree_view_insert_column_with_attributes ( w->tree_effects, EFFECTS_COL_NAME, "name", renderer, "text", EFFECTS_COL_NAME, NULL );

  gtk_tree_selection_set_mode ( gtk_tree_view_get_selection(w->tree_effects), GTK_SELECTION_SINGLE );
  g_signal_connect ( gtk_tree_view_get_selection(w->tree_effects), "changed", G_CALLBACK(on_selected_effect_changed), w );

  g_object_unref ( list_store );

  button_box = GTK_HBOX ( gtk_hbox_new (TRUE,5) );
  gtk_widget_show ( GTK_WIDGET(button_box) );
  gtk_box_pack_start ( box, GTK_WIDGET(button_box), FALSE, TRUE, 0 );

  btn = GTK_BUTTON ( gtk_button_new() );
  gtk_button_set_image ( btn, gtk_image_new_from_stock(GTK_STOCK_ADD,GTK_ICON_SIZE_BUTTON) );
  gtk_widget_show ( GTK_WIDGET(btn) );
  g_signal_connect ( btn, "clicked", G_CALLBACK(on_add_effect_clicked), w );
  gtk_box_pack_start ( GTK_BOX(button_box), GTK_WIDGET(btn), FALSE, FALSE, 0 );

  btn = GTK_BUTTON ( gtk_button_new() );
  gtk_button_set_image ( btn, gtk_image_new_from_stock(GTK_STOCK_REMOVE,GTK_ICON_SIZE_BUTTON) );
  gtk_widget_show ( GTK_WIDGET(btn) );
  g_signal_connect ( btn, "clicked", G_CALLBACK(on_remove_effect_clicked), w );
  gtk_box_pack_start ( GTK_BOX(button_box), GTK_WIDGET(btn), FALSE, FALSE, 0 );

  btn = GTK_BUTTON ( gtk_button_new() );
  gtk_button_set_image ( btn, gtk_image_new_from_stock(GTK_STOCK_CLEAR,GTK_ICON_SIZE_BUTTON) );
  gtk_widget_show ( GTK_WIDGET(btn) );
  g_signal_connect ( btn, "clicked", G_CALLBACK(on_clear_effects_list_clicked), w );
  gtk_box_pack_start ( GTK_BOX(button_box), GTK_WIDGET(btn), FALSE, FALSE, 0 );

  widget = gtk_vseparator_new ( );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(w->hbox_effects), widget, FALSE, TRUE, 0 );
}


/*
 * Show/Hide the UI of a Wave
 */
void
wave_show_ui ( Wave *w )
{
  g_assert ( w != NULL );

  gtk_widget_show ( GTK_WIDGET(w->vbox) );
}

void
wave_hide_ui ( Wave *w )
{
  g_assert ( w != NULL );

  gtk_widget_hide ( GTK_WIDGET(w->vbox) );
}


/*
 * Iterator
 */
gdouble
wave_iter ( Wave *w )
{
  GSList *l;
  Effect *e;
  gdouble val;

  g_assert ( w != NULL );

  if ( w->gen == NULL )
    return 0.0;

  g_mutex_lock ( w->mutex );

  val = w->gen->iter ( w->gen->data );

  for ( l=w->effects; l!=NULL; l=g_slist_next(l) )
    {
      e = (Effect *) l->data;

      if ( e->active )
        val = e->apply ( e->data, val );
    }

  g_mutex_unlock ( w->mutex );

  return val;
}


/*
 * Reset
 */
void
wave_reset ( Wave *w )
{
  GSList *l;
  Effect *e;

  g_assert ( w != NULL );

  g_mutex_lock ( w->mutex );

  if ( w->gen != NULL )
    w->gen->reset (  w->gen->data );

  for ( l=w->effects; l!=NULL; l=g_slist_next(l) )
    {
      e = (Effect *) l->data;
      e->reset ( e->data );
    }

  g_mutex_unlock ( w->mutex );
}


/*
 * Destructor
 */
void
wave_destroy ( Wave *w )
{
  GSList *l;

  g_assert ( w != NULL );

  if ( w->name != NULL )
    g_free ( w->name );

  /* free wave generator */
  if ( w->gen != NULL )
    {
      w->gen->destroy ( w->gen->data );
      g_free ( w->gen );
    }

  /* free effects */
  for ( l=w->effects; l!=NULL; l=g_slist_next(l) )
    effect_destroy ( (Effect *) l->data );

  g_slist_free ( w->effects );

  /* free UI */
  if ( w->vbox != NULL )
    {
      g_hash_table_destroy ( w->generators );
      gtk_widget_destroy ( GTK_WIDGET(w->vbox) );
    }

  g_mutex_free ( w->mutex );

  g_free ( w );
}
