/*
 * Simple Wave Generator
 * Abstract wave generator which provides basic behaviour.
 */

#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include "wave_generator.h"
#include "wave_generator_simple.h"

extern guint SOUND_RATE;

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

/*
 * Initilization of the simple wave generator.
 */
void
wave_generator_simple_init ( gpointer data )
{
  WaveGeneratorSimpleData *wgsd;

  if ( data == NULL )
    return;

  wgsd = (WaveGeneratorSimpleData *) data;

  wgsd->freq = 100.0;
  wgsd->x = 0.0;
  wgsd->x_inc = (wgsd->freq * 2.0 * M_PI) / ((gdouble)SOUND_RATE);

  wgsd->vbox = NULL;
}


/*
 * Cloning operation (ineffective, because this is abstract).
 * Returns a brand new simple wave generator.
 */
gpointer
wave_generator_simple_clone ( gpointer data )
{
  return NULL;
}

/*
 * Generic cloning operation which clones from an instance.
 * Returns a brand new wave generator.
 */
gpointer
wave_generator_simple_clone_from_instance ( gpointer data,
                                            gpointer instance )
{
  WaveGenerator * ret = g_malloc ( sizeof(WaveGenerator) );

  memcpy ( ret, instance, sizeof(WaveGenerator) );

  ret->data = g_malloc ( sizeof(WaveGeneratorSimpleData) );

  if ( data != NULL )
    {
      memcpy ( ret->data, data, sizeof(WaveGeneratorSimpleData) );
      ((WaveGeneratorSimpleData*)ret->data)->vbox = NULL;
    }
  else
    wave_generator_simple_init ( ret->data );

  return ret;
}


/*
 * Makes the configuration user interface
 */
static void
on_freq_changed ( GtkSpinButton *sb,
                  gpointer data )
{
  WaveGeneratorSimpleData *wgsd;

  g_assert ( data != NULL );

  wgsd = (WaveGeneratorSimpleData *) data;
  wgsd->freq = gtk_spin_button_get_value ( sb );
  wgsd->x_inc = (wgsd->freq * 2.0 * M_PI) / ((gdouble)SOUND_RATE);
}

void
wave_generator_simple_make_ui ( gpointer data,
                              GtkBox *parent )
{
  GtkWidget *box;
  GtkWidget *widget;
  WaveGeneratorSimpleData *wgsd;

  g_assert ( data != NULL );

  wgsd = (WaveGeneratorSimpleData *) data;

  wgsd->vbox = GTK_VBOX ( gtk_vbox_new(FALSE,5) );
  gtk_box_pack_start ( parent, GTK_WIDGET(wgsd->vbox), TRUE, TRUE, 0 );

  /* frequency */
  box = gtk_hbox_new ( FALSE, 5 );
  gtk_widget_show ( box );
  gtk_box_pack_start ( GTK_BOX(wgsd->vbox), box, FALSE, FALSE, 0 );

  widget = gtk_label_new ( "frequency (Hz) :" );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, TRUE, 0 );

  widget = gtk_spin_button_new_with_range ( 1.0, 15000.0, 1.0 );
  gtk_spin_button_set_value ( GTK_SPIN_BUTTON(widget), wgsd->freq );
  g_signal_connect ( widget, "value-changed", G_CALLBACK(on_freq_changed), data );
  gtk_widget_show ( widget );
  gtk_box_pack_start ( GTK_BOX(box), widget, FALSE, FALSE, 0 );
}


/*
 * Show/Hide the UI of the wave generator
 */
void
wave_generator_simple_show_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_show ( GTK_WIDGET(((WaveGeneratorSimpleData *)data)->vbox) );
}

void
wave_generator_simple_hide_ui ( gpointer data )
{
  g_assert ( data != NULL );

  gtk_widget_hide ( GTK_WIDGET(((WaveGeneratorSimpleData *)data)->vbox) );
}


/*
 * Iterates the wave generator (useless here as it is abstract).
 * Returns the new value.
 */
gdouble
wave_generator_simple_iter ( gpointer data )
{
  return 0.0;
}


/*
 * Resets the wave generator.
 */
void
wave_generator_simple_reset ( gpointer data )
{
  WaveGeneratorSimpleData *wgsd;

  if ( data == NULL )
    return;

  wgsd = (WaveGeneratorSimpleData *) data;

  wgsd->x = 0.0;
  wgsd->x_inc = (wgsd->freq * 2.0 * M_PI) / ((gdouble)SOUND_RATE);
}


/*
 * Destroys the wave generator.
 */
void
wave_generator_simple_destroy ( gpointer data )
{
  WaveGeneratorSimpleData *wgsd;

  if ( data != NULL )
    {
      wgsd = (WaveGeneratorSimpleData *) data;

      if ( wgsd->vbox != NULL )
        gtk_widget_destroy ( GTK_WIDGET(wgsd->vbox) );

      g_free ( data );
    }
}
